/*  $Id$
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Author: Rafael Sadyrov
 *
 */

#include <ncbi_pch.hpp>
#include "../cgi_impl.hpp"

#define BOOST_AUTO_TEST_MAIN
#include <corelib/test_boost.hpp>

#include <common/test_assert.h>  /* This header must go last */


USING_NCBI_SCOPE;


BOOST_AUTO_TEST_CASE(Test_CCgiEntries_Parser)
{
    enum { eQuery, eFlags, eExpectedEntries, eExpectedIndexes };
    const auto N = CCgiRequest::fIndexesNotEntries;
    const auto S = CCgiRequest::fSemicolonIsNotArgDelimiter;
    const auto I = CCgiRequest::fDisableParsingAsIndex;

    vector<tuple<string, CCgiRequest::TFlags, TCgiEntries, TCgiIndexes>> checks =
    {
        { "",            0,         {                                                             }, {             } },
        { "a=",          0,         { { "a",     ""     },                                        }, {             } },
        { "a=0",         0,         { { "a",     "0"    },                                        }, {             } },
        { "&",           0,         { { "&",     ""     },                                        }, {             } },
        { "a&",          0,         { { "a&",    ""     },                                        }, {             } },
        { "a=&",         0,         { { "a",     ""     },                                        }, {             } },
        { "a=0&",        0,         { { "a",     "0"    },                                        }, {             } },
        { "&b",          0,         {                      { "&b",   ""      },                   }, {             } },
        { "a&b",         0,         { { "a&b",   ""     },                                        }, {             } },
        { "a=&b",        0,         { { "a",     ""     }, { "b",    ""      },                   }, {             } },
        { "a=0&b",       0,         { { "a",     "0"    }, { "b",    ""      },                   }, {             } },
        { "&b=",         0,         {                      { "b",    ""      },                   }, {             } },
        { "a&b=",        0,         { { "a",     ""     }, { "b",    ""      },                   }, {             } },
        { "a=&b=",       0,         { { "a",     ""     }, { "b",    ""      },                   }, {             } },
        { "a=0&b=",      0,         { { "a",     "0"    }, { "b",    ""      },                   }, {             } },
        { "&b=0",        0,         {                      { "b",    "0"     },                   }, {             } },
        { "a&b=0",       0,         { { "a",     ""     }, { "b",    "0"     },                   }, {             } },
        { "a=&b=0",      0,         { { "a",     ""     }, { "b",    "0"     },                   }, {             } },
        { "a=0&b=0",     0,         { { "a",     "0"    }, { "b",    "0"     },                   }, {             } },
        { ";",           0,         {                      { ";",    ""      },                   }, {             } },
        { "b;",          0,         {                      { "b;",   ""      },                   }, {             } },
        { "b=;",         0,         {                      { "b",    ""      },                   }, {             } },
        { "b=0;",        0,         {                      { "b",    "0"     },                   }, {             } },
        { ";c",          0,         {                                           { ";c",  ""     } }, {             } },
        { "b;c",         0,         {                      { "b;c",  ""      },                   }, {             } },
        { "b=;c",        0,         {                      { "b",    ""      }, { "c",   ""     } }, {             } },
        { "b=0;c",       0,         {                      { "b",    "0"     }, { "c",   ""     } }, {             } },
        { ";c=",         0,         {                                           { "c",   ""     } }, {             } },
        { "b;c=",        0,         {                      { "b",    ""      }, { "c",   ""     } }, {             } },
        { "b=;c=",       0,         {                      { "b",    ""      }, { "c",   ""     } }, {             } },
        { "b=0;c=",      0,         {                      { "b",    "0"     }, { "c",   ""     } }, {             } },
        { ";c=0",        0,         {                                           { "c",   "0"    } }, {             } },
        { "b;c=0",       0,         {                      { "b",    ""      }, { "c",   "0"    } }, {             } },
        { "b=;c=0",      0,         {                      { "b",    ""      }, { "c",   "0"    } }, {             } },
        { "b=0;c=0",     0,         {                      { "b",    "0"     }, { "c",   "0"    } }, {             } },
        { "&;",          0,         {                      { "&;",   ""      },                   }, {             } },
        { "&b;",         0,         {                      { "&b;",  ""      },                   }, {             } },
        { "&b=;",        0,         {                      { "b",    ""      },                   }, {             } },
        { "&b=0;",       0,         {                      { "b",    "0"     },                   }, {             } },
        { "&;c",         0,         {                                           { "&;c", ""     } }, {             } },
        { "&b;c",        0,         {                      { "&b;c", ""      },                   }, {             } },
        { "&b=;c",       0,         {                      { "b",    ""      }, { "c",   ""     } }, {             } },
        { "&b=0;c",      0,         {                      { "b",    "0"     }, { "c",   ""     } }, {             } },
        { "&;c=",        0,         {                                           { "c",   ""     } }, {             } },
        { "&b;c=",       0,         {                      { "b",    ""      }, { "c",   ""     } }, {             } },
        { "&b=;c=",      0,         {                      { "b",    ""      }, { "c",   ""     } }, {             } },
        { "&b=0;c=",     0,         {                      { "b",    "0"     }, { "c",   ""     } }, {             } },
        { "&;c=0",       0,         {                                           { "c",   "0"    } }, {             } },
        { "&b;c=0",      0,         {                      { "b",    ""      }, { "c",   "0"    } }, {             } },
        { "&b=;c=0",     0,         {                      { "b",    ""      }, { "c",   "0"    } }, {             } },
        { "&b=0;c=0",    0,         {                      { "b",    "0"     }, { "c",   "0"    } }, {             } },
        { "a&;",         0,         { { "a&;",   ""     },                                        }, {             } },
        { "a&b;",        0,         { { "a&b;",  ""     },                                        }, {             } },
        { "a&b=;",       0,         { { "a",     ""     }, { "b",    ""      },                   }, {             } },
        { "a&b=0;",      0,         { { "a",     ""     }, { "b",    "0"     },                   }, {             } },
        { "a&;c",        0,         { { "a&;c",  ""     },                                        }, {             } },
        { "a&b;c",       0,         { { "a&b;c", ""     },                                        }, {             } },
        { "a&b=;c",      0,         { { "a",     ""     }, { "b",    ""      }, { "c",   ""     } }, {             } },
        { "a&b=0;c",     0,         { { "a",     ""     }, { "b",    "0"     }, { "c",   ""     } }, {             } },
        { "a&;c=",       0,         { { "a",     ""     },                      { "c",   ""     } }, {             } },
        { "a&b;c=",      0,         { { "a",     ""     }, { "b",    ""      }, { "c",   ""     } }, {             } },
        { "a&b=;c=",     0,         { { "a",     ""     }, { "b",    ""      }, { "c",   ""     } }, {             } },
        { "a&b=0;c=",    0,         { { "a",     ""     }, { "b",    "0"     }, { "c",   ""     } }, {             } },
        { "a&;c=0",      0,         { { "a",     ""     },                      { "c",   "0"    } }, {             } },
        { "a&b;c=0",     0,         { { "a",     ""     }, { "b",    ""      }, { "c",   "0"    } }, {             } },
        { "a&b=;c=0",    0,         { { "a",     ""     }, { "b",    ""      }, { "c",   "0"    } }, {             } },
        { "a&b=0;c=0",   0,         { { "a",     ""     }, { "b",    "0"     }, { "c",   "0"    } }, {             } },
        { "a=&;",        0,         { { "a",     ""     },                                        }, {             } },
        { "a=&b;",       0,         { { "a",     ""     }, { "b",    ""      },                   }, {             } },
        { "a=&b=;",      0,         { { "a",     ""     }, { "b",    ""      },                   }, {             } },
        { "a=&b=0;",     0,         { { "a",     ""     }, { "b",    "0"     },                   }, {             } },
        { "a=&;c",       0,         { { "a",     ""     },                      { "c",   ""     } }, {             } },
        { "a=&b;c",      0,         { { "a",     ""     }, { "b",    ""      }, { "c",   ""     } }, {             } },
        { "a=&b=;c",     0,         { { "a",     ""     }, { "b",    ""      }, { "c",   ""     } }, {             } },
        { "a=&b=0;c",    0,         { { "a",     ""     }, { "b",    "0"     }, { "c",   ""     } }, {             } },
        { "a=&;c=",      0,         { { "a",     ""     },                      { "c",   ""     } }, {             } },
        { "a=&b;c=",     0,         { { "a",     ""     }, { "b",    ""      }, { "c",   ""     } }, {             } },
        { "a=&b=;c=",    0,         { { "a",     ""     }, { "b",    ""      }, { "c",   ""     } }, {             } },
        { "a=&b=0;c=",   0,         { { "a",     ""     }, { "b",    "0"     }, { "c",   ""     } }, {             } },
        { "a=&;c=0",     0,         { { "a",     ""     },                      { "c",   "0"    } }, {             } },
        { "a=&b;c=0",    0,         { { "a",     ""     }, { "b",    ""      }, { "c",   "0"    } }, {             } },
        { "a=&b=;c=0",   0,         { { "a",     ""     }, { "b",    ""      }, { "c",   "0"    } }, {             } },
        { "a=&b=0;c=0",  0,         { { "a",     ""     }, { "b",    "0"     }, { "c",   "0"    } }, {             } },
        { "a=0&;",       0,         { { "a",     "0"    },                                        }, {             } },
        { "a=0&b;",      0,         { { "a",     "0"    }, { "b",    ""      },                   }, {             } },
        { "a=0&b=;",     0,         { { "a",     "0"    }, { "b",    ""      },                   }, {             } },
        { "a=0&b=0;",    0,         { { "a",     "0"    }, { "b",    "0"     },                   }, {             } },
        { "a=0&;c",      0,         { { "a",     "0"    },                      { "c",   ""     } }, {             } },
        { "a=0&b;c",     0,         { { "a",     "0"    }, { "b",    ""      }, { "c",   ""     } }, {             } },
        { "a=0&b=;c",    0,         { { "a",     "0"    }, { "b",    ""      }, { "c",   ""     } }, {             } },
        { "a=0&b=0;c",   0,         { { "a",     "0"    }, { "b",    "0"     }, { "c",   ""     } }, {             } },
        { "a=0&;c=",     0,         { { "a",     "0"    },                      { "c",   ""     } }, {             } },
        { "a=0&b;c=",    0,         { { "a",     "0"    }, { "b",    ""      }, { "c",   ""     } }, {             } },
        { "a=0&b=;c=",   0,         { { "a",     "0"    }, { "b",    ""      }, { "c",   ""     } }, {             } },
        { "a=0&b=0;c=",  0,         { { "a",     "0"    }, { "b",    "0"     }, { "c",   ""     } }, {             } },
        { "a=0&;c=0",    0,         { { "a",     "0"    },                      { "c",   "0"    } }, {             } },
        { "a=0&b;c=0",   0,         { { "a",     "0"    }, { "b",    ""      }, { "c",   "0"    } }, {             } },
        { "a=0&b=;c=0",  0,         { { "a",     "0"    }, { "b",    ""      }, { "c",   "0"    } }, {             } },
        { "a=0&b=0;c=0", 0,         { { "a",     "0"    }, { "b",    "0"     }, { "c",   "0"    } }, {             } },
        { "",            N,         {                                                             }, {             } },
        { "a=",          N,         { { "a",     ""     },                                        }, {             } },
        { "a=0",         N,         { { "a",     "0"    },                                        }, {             } },
        { "&",           N,         {                                                             }, { { "&"     } } },
        { "a&",          N,         {                                                             }, { { "a&"    } } },
        { "a=&",         N,         { { "a",     ""     },                                        }, {             } },
        { "a=0&",        N,         { { "a",     "0"    },                                        }, {             } },
        { "&b",          N,         {                                                             }, { { "&b"    } } },
        { "a&b",         N,         {                                                             }, { { "a&b"   } } },
        { "a=&b",        N,         { { "a",     ""     }, { "b",    ""      },                   }, {             } },
        { "a=0&b",       N,         { { "a",     "0"    }, { "b",    ""      },                   }, {             } },
        { "&b=",         N,         {                      { "b",    ""      },                   }, {             } },
        { "a&b=",        N,         { { "a",     ""     }, { "b",    ""      },                   }, {             } },
        { "a=&b=",       N,         { { "a",     ""     }, { "b",    ""      },                   }, {             } },
        { "a=0&b=",      N,         { { "a",     "0"    }, { "b",    ""      },                   }, {             } },
        { "&b=0",        N,         {                      { "b",    "0"     },                   }, {             } },
        { "a&b=0",       N,         { { "a",     ""     }, { "b",    "0"     },                   }, {             } },
        { "a=&b=0",      N,         { { "a",     ""     }, { "b",    "0"     },                   }, {             } },
        { "a=0&b=0",     N,         { { "a",     "0"    }, { "b",    "0"     },                   }, {             } },
        { ";",           N,         {                                                             }, { { ";"     } } },
        { "b;",          N,         {                                                             }, { { "b;"    } } },
        { "b=;",         N,         {                      { "b",    ""      },                   }, {             } },
        { "b=0;",        N,         {                      { "b",    "0"     },                   }, {             } },
        { ";c",          N,         {                                                             }, { { ";c"    } } },
        { "b;c",         N,         {                                                             }, { { "b;c"   } } },
        { "b=;c",        N,         {                      { "b",    ""      }, { "c",   ""     } }, {             } },
        { "b=0;c",       N,         {                      { "b",    "0"     }, { "c",   ""     } }, {             } },
        { ";c=",         N,         {                                           { "c",   ""     } }, {             } },
        { "b;c=",        N,         {                      { "b",    ""      }, { "c",   ""     } }, {             } },
        { "b=;c=",       N,         {                      { "b",    ""      }, { "c",   ""     } }, {             } },
        { "b=0;c=",      N,         {                      { "b",    "0"     }, { "c",   ""     } }, {             } },
        { ";c=0",        N,         {                                           { "c",   "0"    } }, {             } },
        { "b;c=0",       N,         {                      { "b",    ""      }, { "c",   "0"    } }, {             } },
        { "b=;c=0",      N,         {                      { "b",    ""      }, { "c",   "0"    } }, {             } },
        { "b=0;c=0",     N,         {                      { "b",    "0"     }, { "c",   "0"    } }, {             } },
        { "&;",          N,         {                                                             }, { { "&;"    } } },
        { "&b;",         N,         {                                                             }, { { "&b;"   } } },
        { "&b=;",        N,         {                      { "b",    ""      },                   }, {             } },
        { "&b=0;",       N,         {                      { "b",    "0"     },                   }, {             } },
        { "&;c",         N,         {                                                             }, { { "&;c"   } } },
        { "&b;c",        N,         {                                                             }, { { "&b;c"  } } },
        { "&b=;c",       N,         {                      { "b",    ""      }, { "c",   ""     } }, {             } },
        { "&b=0;c",      N,         {                      { "b",    "0"     }, { "c",   ""     } }, {             } },
        { "&;c=",        N,         {                                           { "c",   ""     } }, {             } },
        { "&b;c=",       N,         {                      { "b",    ""      }, { "c",   ""     } }, {             } },
        { "&b=;c=",      N,         {                      { "b",    ""      }, { "c",   ""     } }, {             } },
        { "&b=0;c=",     N,         {                      { "b",    "0"     }, { "c",   ""     } }, {             } },
        { "&;c=0",       N,         {                                           { "c",   "0"    } }, {             } },
        { "&b;c=0",      N,         {                      { "b",    ""      }, { "c",   "0"    } }, {             } },
        { "&b=;c=0",     N,         {                      { "b",    ""      }, { "c",   "0"    } }, {             } },
        { "&b=0;c=0",    N,         {                      { "b",    "0"     }, { "c",   "0"    } }, {             } },
        { "a&;",         N,         {                                                             }, { { "a&;"   } } },
        { "a&b;",        N,         {                                                             }, { { "a&b;"  } } },
        { "a&b=;",       N,         { { "a",     ""     }, { "b",    ""      },                   }, {             } },
        { "a&b=0;",      N,         { { "a",     ""     }, { "b",    "0"     },                   }, {             } },
        { "a&;c",        N,         {                                                             }, { { "a&;c"  } } },
        { "a&b;c",       N,         {                                                             }, { { "a&b;c" } } },
        { "a&b=;c",      N,         { { "a",     ""     }, { "b",    ""      }, { "c",   ""     } }, {             } },
        { "a&b=0;c",     N,         { { "a",     ""     }, { "b",    "0"     }, { "c",   ""     } }, {             } },
        { "a&;c=",       N,         { { "a",     ""     },                      { "c",   ""     } }, {             } },
        { "a&b;c=",      N,         { { "a",     ""     }, { "b",    ""      }, { "c",   ""     } }, {             } },
        { "a&b=;c=",     N,         { { "a",     ""     }, { "b",    ""      }, { "c",   ""     } }, {             } },
        { "a&b=0;c=",    N,         { { "a",     ""     }, { "b",    "0"     }, { "c",   ""     } }, {             } },
        { "a&;c=0",      N,         { { "a",     ""     },                      { "c",   "0"    } }, {             } },
        { "a&b;c=0",     N,         { { "a",     ""     }, { "b",    ""      }, { "c",   "0"    } }, {             } },
        { "a&b=;c=0",    N,         { { "a",     ""     }, { "b",    ""      }, { "c",   "0"    } }, {             } },
        { "a&b=0;c=0",   N,         { { "a",     ""     }, { "b",    "0"     }, { "c",   "0"    } }, {             } },
        { "a=&;",        N,         { { "a",     ""     },                                        }, {             } },
        { "a=&b;",       N,         { { "a",     ""     }, { "b",    ""      },                   }, {             } },
        { "a=&b=;",      N,         { { "a",     ""     }, { "b",    ""      },                   }, {             } },
        { "a=&b=0;",     N,         { { "a",     ""     }, { "b",    "0"     },                   }, {             } },
        { "a=&;c",       N,         { { "a",     ""     },                      { "c",   ""     } }, {             } },
        { "a=&b;c",      N,         { { "a",     ""     }, { "b",    ""      }, { "c",   ""     } }, {             } },
        { "a=&b=;c",     N,         { { "a",     ""     }, { "b",    ""      }, { "c",   ""     } }, {             } },
        { "a=&b=0;c",    N,         { { "a",     ""     }, { "b",    "0"     }, { "c",   ""     } }, {             } },
        { "a=&;c=",      N,         { { "a",     ""     },                      { "c",   ""     } }, {             } },
        { "a=&b;c=",     N,         { { "a",     ""     }, { "b",    ""      }, { "c",   ""     } }, {             } },
        { "a=&b=;c=",    N,         { { "a",     ""     }, { "b",    ""      }, { "c",   ""     } }, {             } },
        { "a=&b=0;c=",   N,         { { "a",     ""     }, { "b",    "0"     }, { "c",   ""     } }, {             } },
        { "a=&;c=0",     N,         { { "a",     ""     },                      { "c",   "0"    } }, {             } },
        { "a=&b;c=0",    N,         { { "a",     ""     }, { "b",    ""      }, { "c",   "0"    } }, {             } },
        { "a=&b=;c=0",   N,         { { "a",     ""     }, { "b",    ""      }, { "c",   "0"    } }, {             } },
        { "a=&b=0;c=0",  N,         { { "a",     ""     }, { "b",    "0"     }, { "c",   "0"    } }, {             } },
        { "a=0&;",       N,         { { "a",     "0"    },                                        }, {             } },
        { "a=0&b;",      N,         { { "a",     "0"    }, { "b",    ""      },                   }, {             } },
        { "a=0&b=;",     N,         { { "a",     "0"    }, { "b",    ""      },                   }, {             } },
        { "a=0&b=0;",    N,         { { "a",     "0"    }, { "b",    "0"     },                   }, {             } },
        { "a=0&;c",      N,         { { "a",     "0"    },                      { "c",   ""     } }, {             } },
        { "a=0&b;c",     N,         { { "a",     "0"    }, { "b",    ""      }, { "c",   ""     } }, {             } },
        { "a=0&b=;c",    N,         { { "a",     "0"    }, { "b",    ""      }, { "c",   ""     } }, {             } },
        { "a=0&b=0;c",   N,         { { "a",     "0"    }, { "b",    "0"     }, { "c",   ""     } }, {             } },
        { "a=0&;c=",     N,         { { "a",     "0"    },                      { "c",   ""     } }, {             } },
        { "a=0&b;c=",    N,         { { "a",     "0"    }, { "b",    ""      }, { "c",   ""     } }, {             } },
        { "a=0&b=;c=",   N,         { { "a",     "0"    }, { "b",    ""      }, { "c",   ""     } }, {             } },
        { "a=0&b=0;c=",  N,         { { "a",     "0"    }, { "b",    "0"     }, { "c",   ""     } }, {             } },
        { "a=0&;c=0",    N,         { { "a",     "0"    },                      { "c",   "0"    } }, {             } },
        { "a=0&b;c=0",   N,         { { "a",     "0"    }, { "b",    ""      }, { "c",   "0"    } }, {             } },
        { "a=0&b=;c=0",  N,         { { "a",     "0"    }, { "b",    ""      }, { "c",   "0"    } }, {             } },
        { "a=0&b=0;c=0", N,         { { "a",     "0"    }, { "b",    "0"     }, { "c",   "0"    } }, {             } },
        { "",            S,         {                                                             }, {             } },
        { "a=",          S,         { { "a",     ""     },                                        }, {             } },
        { "a=0",         S,         { { "a",     "0"    },                                        }, {             } },
        { "&",           S,         { { "&",     ""     },                                        }, {             } },
        { "a&",          S,         { { "a&",    ""     },                                        }, {             } },
        { "a=&",         S,         { { "a",     ""     },                                        }, {             } },
        { "a=0&",        S,         { { "a",     "0"    },                                        }, {             } },
        { "&b",          S,         {                      { "&b",   ""      },                   }, {             } },
        { "a&b",         S,         { { "a&b",   ""     },                                        }, {             } },
        { "a=&b",        S,         { { "a",     ""     }, { "b",    ""      },                   }, {             } },
        { "a=0&b",       S,         { { "a",     "0"    }, { "b",    ""      },                   }, {             } },
        { "&b=",         S,         {                      { "b",    ""      },                   }, {             } },
        { "a&b=",        S,         { { "a",     ""     }, { "b",    ""      },                   }, {             } },
        { "a=&b=",       S,         { { "a",     ""     }, { "b",    ""      },                   }, {             } },
        { "a=0&b=",      S,         { { "a",     "0"    }, { "b",    ""      },                   }, {             } },
        { "&b=0",        S,         {                      { "b",    "0"     },                   }, {             } },
        { "a&b=0",       S,         { { "a",     ""     }, { "b",    "0"     },                   }, {             } },
        { "a=&b=0",      S,         { { "a",     ""     }, { "b",    "0"     },                   }, {             } },
        { "a=0&b=0",     S,         { { "a",     "0"    }, { "b",    "0"     },                   }, {             } },
        { ";",           S,         {                      { ";",    ""      },                   }, {             } },
        { "b;",          S,         {                      { "b;",   ""      },                   }, {             } },
        { "b=;",         S,         {                      { "b",    ";"     },                   }, {             } },
        { "b=0;",        S,         {                      { "b",    "0;"    },                   }, {             } },
        { ";c",          S,         {                                           { ";c",  ""     } }, {             } },
        { "b;c",         S,         {                      { "b;c",  ""      },                   }, {             } },
        { "b=;c",        S,         {                      { "b",    ";c"    },                   }, {             } },
        { "b=0;c",       S,         {                      { "b",    "0;c"   },                   }, {             } },
        { ";c=",         S,         {                                           { ";c",  ""     } }, {             } },
        { "b;c=",        S,         {                      { "b;c",  ""      }                    }, {             } },
        { "b=;c=",       S,         {                      { "b",    ";c="   },                   }, {             } },
        { "b=0;c=",      S,         {                      { "b",    "0;c="  },                   }, {             } },
        { ";c=0",        S,         {                                           { ";c",  "0"    } }, {             } },
        { "b;c=0",       S,         {                      { "b;c",  "0"     },                   }, {             } },
        { "b=;c=0",      S,         {                      { "b",    ";c=0"  },                   }, {             } },
        { "b=0;c=0",     S,         {                      { "b",    "0;c=0" },                   }, {             } },
        { "&;",          S,         {                      { "&;",   ""      },                   }, {             } },
        { "&b;",         S,         {                      { "&b;",  ""      },                   }, {             } },
        { "&b=;",        S,         {                      { "b",    ";"     },                   }, {             } },
        { "&b=0;",       S,         {                      { "b",    "0;"    },                   }, {             } },
        { "&;c",         S,         {                                           { "&;c", ""     } }, {             } },
        { "&b;c",        S,         {                      { "&b;c", ""      },                   }, {             } },
        { "&b=;c",       S,         {                      { "b",    ";c"    },                   }, {             } },
        { "&b=0;c",      S,         {                      { "b",    "0;c"   },                   }, {             } },
        { "&;c=",        S,         {                                           { ";c",  ""     } }, {             } },
        { "&b;c=",       S,         {                      { "b;c",   ""     },                   }, {             } },
        { "&b=;c=",      S,         {                      { "b",    ";c="   },                   }, {             } },
        { "&b=0;c=",     S,         {                      { "b",    "0;c="  },                   }, {             } },
        { "&;c=0",       S,         {                                           { ";c",  "0"    } }, {             } },
        { "&b;c=0",      S,         {                      { "b;c",  "0"     },                   }, {             } },
        { "&b=;c=0",     S,         {                      { "b",    ";c=0"  },                   }, {             } },
        { "&b=0;c=0",    S,         {                      { "b",    "0;c=0" },                   }, {             } },
        { "a&;",         S,         { { "a&;",   ""     },                                        }, {             } },
        { "a&b;",        S,         { { "a&b;",  ""     },                                        }, {             } },
        { "a&b=;",       S,         { { "a",     ""     }, { "b",    ";"     },                   }, {             } },
        { "a&b=0;",      S,         { { "a",     ""     }, { "b",    "0;"    },                   }, {             } },
        { "a&;c",        S,         { { "a&;c",  ""     },                                        }, {             } },
        { "a&b;c",       S,         { { "a&b;c", ""     },                                        }, {             } },
        { "a&b=;c",      S,         { { "a",     ""     }, { "b",    ";c"    },                   }, {             } },
        { "a&b=0;c",     S,         { { "a",     ""     }, { "b",    "0;c"   },                   }, {             } },
        { "a&;c=",       S,         { { "a",     ""     },                      { ";c",  ""     } }, {             } },
        { "a&b;c=",      S,         { { "a",     ""     }, { "b;c",  ""      },                   }, {             } },
        { "a&b=;c=",     S,         { { "a",     ""     }, { "b",    ";c="   },                   }, {             } },
        { "a&b=0;c=",    S,         { { "a",     ""     }, { "b",    "0;c="  },                   }, {             } },
        { "a&;c=0",      S,         { { "a",     ""     },                      { ";c",  "0"    } }, {             } },
        { "a&b;c=0",     S,         { { "a",     ""     }, { "b;c",  "0"     },                   }, {             } },
        { "a&b=;c=0",    S,         { { "a",     ""     }, { "b",    ";c=0"  },                   }, {             } },
        { "a&b=0;c=0",   S,         { { "a",     ""     }, { "b",    "0;c=0" },                   }, {             } },
        { "a=&;",        S,         { { "a",     ""     }, { ";",    ""      },                   }, {             } },
        { "a=&b;",       S,         { { "a",     ""     }, { "b;",   ""      },                   }, {             } },
        { "a=&b=;",      S,         { { "a",     ""     }, { "b",    ";"     },                   }, {             } },
        { "a=&b=0;",     S,         { { "a",     ""     }, { "b",    "0;"    },                   }, {             } },
        { "a=&;c",       S,         { { "a",     ""     },                      { ";c",  ""     } }, {             } },
        { "a=&b;c",      S,         { { "a",     ""     }, { "b;c",  ""      },                   }, {             } },
        { "a=&b=;c",     S,         { { "a",     ""     }, { "b",    ";c"    },                   }, {             } },
        { "a=&b=0;c",    S,         { { "a",     ""     }, { "b",    "0;c"   },                   }, {             } },
        { "a=&;c=",      S,         { { "a",     ""     },                      { ";c",  ""     } }, {             } },
        { "a=&b;c=",     S,         { { "a",     ""     }, { "b;c",  ""      },                   }, {             } },
        { "a=&b=;c=",    S,         { { "a",     ""     }, { "b",    ";c="   },                   }, {             } },
        { "a=&b=0;c=",   S,         { { "a",     ""     }, { "b",    "0;c="  },                   }, {             } },
        { "a=&;c=0",     S,         { { "a",     ""     },                      { ";c",  "0"    } }, {             } },
        { "a=&b;c=0",    S,         { { "a",     ""     }, { "b;c",  "0"     },                   }, {             } },
        { "a=&b=;c=0",   S,         { { "a",     ""     }, { "b",    ";c=0"  },                   }, {             } },
        { "a=&b=0;c=0",  S,         { { "a",     ""     }, { "b",    "0;c=0" },                   }, {             } },
        { "a=0&;",       S,         { { "a",     "0"    }, { ";",    ""      },                   }, {             } },
        { "a=0&b;",      S,         { { "a",     "0"    }, { "b;",   ""      },                   }, {             } },
        { "a=0&b=;",     S,         { { "a",     "0"    }, { "b",    ";"     },                   }, {             } },
        { "a=0&b=0;",    S,         { { "a",     "0"    }, { "b",    "0;"    },                   }, {             } },
        { "a=0&;c",      S,         { { "a",     "0"    },                      { ";c",  ""     } }, {             } },
        { "a=0&b;c",     S,         { { "a",     "0"    }, { "b;c",  ""      },                   }, {             } },
        { "a=0&b=;c",    S,         { { "a",     "0"    }, { "b",    ";c"    },                   }, {             } },
        { "a=0&b=0;c",   S,         { { "a",     "0"    }, { "b",    "0;c"   },                   }, {             } },
        { "a=0&;c=",     S,         { { "a",     "0"    },                      { ";c",  ""     } }, {             } },
        { "a=0&b;c=",    S,         { { "a",     "0"    }, { "b;c",  ""      },                   }, {             } },
        { "a=0&b=;c=",   S,         { { "a",     "0"    }, { "b",    ";c="   },                   }, {             } },
        { "a=0&b=0;c=",  S,         { { "a",     "0"    }, { "b",    "0;c="  },                   }, {             } },
        { "a=0&;c=0",    S,         { { "a",     "0"    },                      { ";c",  "0"    } }, {             } },
        { "a=0&b;c=0",   S,         { { "a",     "0"    }, { "b;c",  "0"     },                   }, {             } },
        { "a=0&b=;c=0",  S,         { { "a",     "0"    }, { "b",    ";c=0"  },                   }, {             } },
        { "a=0&b=0;c=0", S,         { { "a",     "0"    }, { "b",    "0;c=0" },                   }, {             } },
        { "",            N | S,     {                                                             }, {             } },
        { "a=",          N | S,     { { "a",     ""     },                                        }, {             } },
        { "a=0",         N | S,     { { "a",     "0"    },                                        }, {             } },
        { "&",           N | S,     {                                                             }, { { "&"     } } },
        { "a&",          N | S,     {                                                             }, { { "a&"    } } },
        { "a=&",         N | S,     { { "a",     ""     },                                        }, {             } },
        { "a=0&",        N | S,     { { "a",     "0"    },                                        }, {             } },
        { "&b",          N | S,     {                                                             }, { { "&b"    } } },
        { "a&b",         N | S,     {                                                             }, { { "a&b"   } } },
        { "a=&b",        N | S,     { { "a",     ""     }, { "b",    ""      },                   }, {             } },
        { "a=0&b",       N | S,     { { "a",     "0"    }, { "b",    ""      },                   }, {             } },
        { "&b=",         N | S,     {                      { "b",    ""      },                   }, {             } },
        { "a&b=",        N | S,     { { "a",     ""     }, { "b",    ""      },                   }, {             } },
        { "a=&b=",       N | S,     { { "a",     ""     }, { "b",    ""      },                   }, {             } },
        { "a=0&b=",      N | S,     { { "a",     "0"    }, { "b",    ""      },                   }, {             } },
        { "&b=0",        N | S,     {                      { "b",    "0"     },                   }, {             } },
        { "a&b=0",       N | S,     { { "a",     ""     }, { "b",    "0"     },                   }, {             } },
        { "a=&b=0",      N | S,     { { "a",     ""     }, { "b",    "0"     },                   }, {             } },
        { "a=0&b=0",     N | S,     { { "a",     "0"    }, { "b",    "0"     },                   }, {             } },
        { ";",           N | S,     {                                                             }, { { ";"     } } },
        { "b;",          N | S,     {                                                             }, { { "b;"    } } },
        { "b=;",         N | S,     {                      { "b",    ";"     },                   }, {             } },
        { "b=0;",        N | S,     {                      { "b",    "0;"    },                   }, {             } },
        { ";c",          N | S,     {                                                             }, { { ";c"    } } },
        { "b;c",         N | S,     {                                                             }, { { "b;c"   } } },
        { "b=;c",        N | S,     {                      { "b",    ";c"    },                   }, {             } },
        { "b=0;c",       N | S,     {                      { "b",    "0;c"   },                   }, {             } },
        { ";c=",         N | S,     {                                           { ";c",  ""     } }, {             } },
        { "b;c=",        N | S,     {                      { "b;c",  ""      }                    }, {             } },
        { "b=;c=",       N | S,     {                      { "b",    ";c="   },                   }, {             } },
        { "b=0;c=",      N | S,     {                      { "b",    "0;c="  },                   }, {             } },
        { ";c=0",        N | S,     {                                           { ";c",  "0"    } }, {             } },
        { "b;c=0",       N | S,     {                      { "b;c",  "0"     },                   }, {             } },
        { "b=;c=0",      N | S,     {                      { "b",    ";c=0"  },                   }, {             } },
        { "b=0;c=0",     N | S,     {                      { "b",    "0;c=0" },                   }, {             } },
        { "&;",          N | S,     {                                                             }, { { "&;"    } } },
        { "&b;",         N | S,     {                                                             }, { { "&b;"   } } },
        { "&b=;",        N | S,     {                      { "b",    ";"     },                   }, {             } },
        { "&b=0;",       N | S,     {                      { "b",    "0;"    },                   }, {             } },
        { "&;c",         N | S,     {                                                             }, { { "&;c"   } } },
        { "&b;c",        N | S,     {                                                             }, { { "&b;c"  } } },
        { "&b=;c",       N | S,     {                      { "b",    ";c"    },                   }, {             } },
        { "&b=0;c",      N | S,     {                      { "b",    "0;c"   },                   }, {             } },
        { "&;c=",        N | S,     {                                           { ";c",  ""     } }, {             } },
        { "&b;c=",       N | S,     {                      { "b;c",   ""     },                   }, {             } },
        { "&b=;c=",      N | S,     {                      { "b",    ";c="   },                   }, {             } },
        { "&b=0;c=",     N | S,     {                      { "b",    "0;c="  },                   }, {             } },
        { "&;c=0",       N | S,     {                                           { ";c",  "0"    } }, {             } },
        { "&b;c=0",      N | S,     {                      { "b;c",  "0"     },                   }, {             } },
        { "&b=;c=0",     N | S,     {                      { "b",    ";c=0"  },                   }, {             } },
        { "&b=0;c=0",    N | S,     {                      { "b",    "0;c=0" },                   }, {             } },
        { "a&;",         N | S,     {                                                             }, { { "a&;"   } } },
        { "a&b;",        N | S,     {                                                             }, { { "a&b;"  } } },
        { "a&b=;",       N | S,     { { "a",     ""     }, { "b",    ";"     },                   }, {             } },
        { "a&b=0;",      N | S,     { { "a",     ""     }, { "b",    "0;"    },                   }, {             } },
        { "a&;c",        N | S,     {                                                             }, { { "a&;c"  } } },
        { "a&b;c",       N | S,     {                                                             }, { { "a&b;c" } } },
        { "a&b=;c",      N | S,     { { "a",     ""     }, { "b",    ";c"    },                   }, {             } },
        { "a&b=0;c",     N | S,     { { "a",     ""     }, { "b",    "0;c"   },                   }, {             } },
        { "a&;c=",       N | S,     { { "a",     ""     },                      { ";c",  ""     } }, {             } },
        { "a&b;c=",      N | S,     { { "a",     ""     }, { "b;c",  ""      },                   }, {             } },
        { "a&b=;c=",     N | S,     { { "a",     ""     }, { "b",    ";c="   },                   }, {             } },
        { "a&b=0;c=",    N | S,     { { "a",     ""     }, { "b",    "0;c="  },                   }, {             } },
        { "a&;c=0",      N | S,     { { "a",     ""     },                      { ";c",  "0"    } }, {             } },
        { "a&b;c=0",     N | S,     { { "a",     ""     }, { "b;c",  "0"     },                   }, {             } },
        { "a&b=;c=0",    N | S,     { { "a",     ""     }, { "b",    ";c=0"  },                   }, {             } },
        { "a&b=0;c=0",   N | S,     { { "a",     ""     }, { "b",    "0;c=0" },                   }, {             } },
        { "a=&;",        N | S,     { { "a",     ""     }, { ";",    ""      },                   }, {             } },
        { "a=&b;",       N | S,     { { "a",     ""     }, { "b;",   ""      },                   }, {             } },
        { "a=&b=;",      N | S,     { { "a",     ""     }, { "b",    ";"     },                   }, {             } },
        { "a=&b=0;",     N | S,     { { "a",     ""     }, { "b",    "0;"    },                   }, {             } },
        { "a=&;c",       N | S,     { { "a",     ""     },                      { ";c",  ""     } }, {             } },
        { "a=&b;c",      N | S,     { { "a",     ""     }, { "b;c",  ""      },                   }, {             } },
        { "a=&b=;c",     N | S,     { { "a",     ""     }, { "b",    ";c"    },                   }, {             } },
        { "a=&b=0;c",    N | S,     { { "a",     ""     }, { "b",    "0;c"   },                   }, {             } },
        { "a=&;c=",      N | S,     { { "a",     ""     },                      { ";c",  ""     } }, {             } },
        { "a=&b;c=",     N | S,     { { "a",     ""     }, { "b;c",  ""      },                   }, {             } },
        { "a=&b=;c=",    N | S,     { { "a",     ""     }, { "b",    ";c="   },                   }, {             } },
        { "a=&b=0;c=",   N | S,     { { "a",     ""     }, { "b",    "0;c="  },                   }, {             } },
        { "a=&;c=0",     N | S,     { { "a",     ""     },                      { ";c",  "0"    } }, {             } },
        { "a=&b;c=0",    N | S,     { { "a",     ""     }, { "b;c",  "0"     },                   }, {             } },
        { "a=&b=;c=0",   N | S,     { { "a",     ""     }, { "b",    ";c=0"  },                   }, {             } },
        { "a=&b=0;c=0",  N | S,     { { "a",     ""     }, { "b",    "0;c=0" },                   }, {             } },
        { "a=0&;",       N | S,     { { "a",     "0"    }, { ";",    ""      },                   }, {             } },
        { "a=0&b;",      N | S,     { { "a",     "0"    }, { "b;",   ""      },                   }, {             } },
        { "a=0&b=;",     N | S,     { { "a",     "0"    }, { "b",    ";"     },                   }, {             } },
        { "a=0&b=0;",    N | S,     { { "a",     "0"    }, { "b",    "0;"    },                   }, {             } },
        { "a=0&;c",      N | S,     { { "a",     "0"    },                      { ";c",  ""     } }, {             } },
        { "a=0&b;c",     N | S,     { { "a",     "0"    }, { "b;c",  ""      },                   }, {             } },
        { "a=0&b=;c",    N | S,     { { "a",     "0"    }, { "b",    ";c"    },                   }, {             } },
        { "a=0&b=0;c",   N | S,     { { "a",     "0"    }, { "b",    "0;c"   },                   }, {             } },
        { "a=0&;c=",     N | S,     { { "a",     "0"    },                      { ";c",  ""     } }, {             } },
        { "a=0&b;c=",    N | S,     { { "a",     "0"    }, { "b;c",  ""      },                   }, {             } },
        { "a=0&b=;c=",   N | S,     { { "a",     "0"    }, { "b",    ";c="   },                   }, {             } },
        { "a=0&b=0;c=",  N | S,     { { "a",     "0"    }, { "b",    "0;c="  },                   }, {             } },
        { "a=0&;c=0",    N | S,     { { "a",     "0"    },                      { ";c",  "0"    } }, {             } },
        { "a=0&b;c=0",   N | S,     { { "a",     "0"    }, { "b;c",  "0"     },                   }, {             } },
        { "a=0&b=;c=0",  N | S,     { { "a",     "0"    }, { "b",    ";c=0"  },                   }, {             } },
        { "a=0&b=0;c=0", N | S,     { { "a",     "0"    }, { "b",    "0;c=0" },                   }, {             } },
        { "",            I,         {                                                             }, {             } },
        { "a=",          I,         { { "a",     ""     },                                        }, {             } },
        { "a=0",         I,         { { "a",     "0"    },                                        }, {             } },
        { "&",           I,         {                                                             }, {             } },
        { "a&",          I,         { { "a",     ""     },                                        }, {             } },
        { "a=&",         I,         { { "a",     ""     },                                        }, {             } },
        { "a=0&",        I,         { { "a",     "0"    },                                        }, {             } },
        { "&b",          I,         {                      { "b",    ""      },                   }, {             } },
        { "a&b",         I,         { { "a",     ""     }, { "b",    ""      },                   }, {             } },
        { "a=&b",        I,         { { "a",     ""     }, { "b",    ""      },                   }, {             } },
        { "a=0&b",       I,         { { "a",     "0"    }, { "b",    ""      },                   }, {             } },
        { "&b=",         I,         {                      { "b",    ""      },                   }, {             } },
        { "a&b=",        I,         { { "a",     ""     }, { "b",    ""      },                   }, {             } },
        { "a=&b=",       I,         { { "a",     ""     }, { "b",    ""      },                   }, {             } },
        { "a=0&b=",      I,         { { "a",     "0"    }, { "b",    ""      },                   }, {             } },
        { "&b=0",        I,         {                      { "b",    "0"     },                   }, {             } },
        { "a&b=0",       I,         { { "a",     ""     }, { "b",    "0"     },                   }, {             } },
        { "a=&b=0",      I,         { { "a",     ""     }, { "b",    "0"     },                   }, {             } },
        { "a=0&b=0",     I,         { { "a",     "0"    }, { "b",    "0"     },                   }, {             } },
        { ";",           I,         {                                                             }, {             } },
        { "b;",          I,         {                      { "b",    ""      },                   }, {             } },
        { "b=;",         I,         {                      { "b",    ""      },                   }, {             } },
        { "b=0;",        I,         {                      { "b",    "0"     },                   }, {             } },
        { ";c",          I,         {                                           { "c",   ""     } }, {             } },
        { "b;c",         I,         {                      { "b",    ""      }, { "c",   ""     } }, {             } },
        { "b=;c",        I,         {                      { "b",    ""      }, { "c",   ""     } }, {             } },
        { "b=0;c",       I,         {                      { "b",    "0"     }, { "c",   ""     } }, {             } },
        { ";c=",         I,         {                                           { "c",   ""     } }, {             } },
        { "b;c=",        I,         {                      { "b",    ""      }, { "c",   ""     } }, {             } },
        { "b=;c=",       I,         {                      { "b",    ""      }, { "c",   ""     } }, {             } },
        { "b=0;c=",      I,         {                      { "b",    "0"     }, { "c",   ""     } }, {             } },
        { ";c=0",        I,         {                                           { "c",   "0"    } }, {             } },
        { "b;c=0",       I,         {                      { "b",    ""      }, { "c",   "0"    } }, {             } },
        { "b=;c=0",      I,         {                      { "b",    ""      }, { "c",   "0"    } }, {             } },
        { "b=0;c=0",     I,         {                      { "b",    "0"     }, { "c",   "0"    } }, {             } },
        { "&;",          I,         {                                                             }, {             } },
        { "&b;",         I,         {                      { "b",    ""      },                   }, {             } },
        { "&b=;",        I,         {                      { "b",    ""      },                   }, {             } },
        { "&b=0;",       I,         {                      { "b",    "0"     },                   }, {             } },
        { "&;c",         I,         {                                           { "c",   ""     } }, {             } },
        { "&b;c",        I,         {                      { "b",    ""      }, { "c",   ""     } }, {             } },
        { "&b=;c",       I,         {                      { "b",    ""      }, { "c",   ""     } }, {             } },
        { "&b=0;c",      I,         {                      { "b",    "0"     }, { "c",   ""     } }, {             } },
        { "&;c=",        I,         {                                           { "c",   ""     } }, {             } },
        { "&b;c=",       I,         {                      { "b",    ""      }, { "c",   ""     } }, {             } },
        { "&b=;c=",      I,         {                      { "b",    ""      }, { "c",   ""     } }, {             } },
        { "&b=0;c=",     I,         {                      { "b",    "0"     }, { "c",   ""     } }, {             } },
        { "&;c=0",       I,         {                                           { "c",   "0"    } }, {             } },
        { "&b;c=0",      I,         {                      { "b",    ""      }, { "c",   "0"    } }, {             } },
        { "&b=;c=0",     I,         {                      { "b",    ""      }, { "c",   "0"    } }, {             } },
        { "&b=0;c=0",    I,         {                      { "b",    "0"     }, { "c",   "0"    } }, {             } },
        { "a&;",         I,         { { "a",     ""     },                                        }, {             } },
        { "a&b;",        I,         { { "a",     ""     }, { "b",    ""      },                   }, {             } },
        { "a&b=;",       I,         { { "a",     ""     }, { "b",    ""      },                   }, {             } },
        { "a&b=0;",      I,         { { "a",     ""     }, { "b",    "0"     },                   }, {             } },
        { "a&;c",        I,         { { "a",     ""     },                      { "c",   ""     } }, {             } },
        { "a&b;c",       I,         { { "a",     ""     }, { "b",    ""      }, { "c",   ""     } }, {             } },
        { "a&b=;c",      I,         { { "a",     ""     }, { "b",    ""      }, { "c",   ""     } }, {             } },
        { "a&b=0;c",     I,         { { "a",     ""     }, { "b",    "0"     }, { "c",   ""     } }, {             } },
        { "a&;c=",       I,         { { "a",     ""     },                      { "c",   ""     } }, {             } },
        { "a&b;c=",      I,         { { "a",     ""     }, { "b",    ""      }, { "c",   ""     } }, {             } },
        { "a&b=;c=",     I,         { { "a",     ""     }, { "b",    ""      }, { "c",   ""     } }, {             } },
        { "a&b=0;c=",    I,         { { "a",     ""     }, { "b",    "0"     }, { "c",   ""     } }, {             } },
        { "a&;c=0",      I,         { { "a",     ""     },                      { "c",   "0"    } }, {             } },
        { "a&b;c=0",     I,         { { "a",     ""     }, { "b",    ""      }, { "c",   "0"    } }, {             } },
        { "a&b=;c=0",    I,         { { "a",     ""     }, { "b",    ""      }, { "c",   "0"    } }, {             } },
        { "a&b=0;c=0",   I,         { { "a",     ""     }, { "b",    "0"     }, { "c",   "0"    } }, {             } },
        { "a=&;",        I,         { { "a",     ""     },                                        }, {             } },
        { "a=&b;",       I,         { { "a",     ""     }, { "b",    ""      },                   }, {             } },
        { "a=&b=;",      I,         { { "a",     ""     }, { "b",    ""      },                   }, {             } },
        { "a=&b=0;",     I,         { { "a",     ""     }, { "b",    "0"     },                   }, {             } },
        { "a=&;c",       I,         { { "a",     ""     },                      { "c",   ""     } }, {             } },
        { "a=&b;c",      I,         { { "a",     ""     }, { "b",    ""      }, { "c",   ""     } }, {             } },
        { "a=&b=;c",     I,         { { "a",     ""     }, { "b",    ""      }, { "c",   ""     } }, {             } },
        { "a=&b=0;c",    I,         { { "a",     ""     }, { "b",    "0"     }, { "c",   ""     } }, {             } },
        { "a=&;c=",      I,         { { "a",     ""     },                      { "c",   ""     } }, {             } },
        { "a=&b;c=",     I,         { { "a",     ""     }, { "b",    ""      }, { "c",   ""     } }, {             } },
        { "a=&b=;c=",    I,         { { "a",     ""     }, { "b",    ""      }, { "c",   ""     } }, {             } },
        { "a=&b=0;c=",   I,         { { "a",     ""     }, { "b",    "0"     }, { "c",   ""     } }, {             } },
        { "a=&;c=0",     I,         { { "a",     ""     },                      { "c",   "0"    } }, {             } },
        { "a=&b;c=0",    I,         { { "a",     ""     }, { "b",    ""      }, { "c",   "0"    } }, {             } },
        { "a=&b=;c=0",   I,         { { "a",     ""     }, { "b",    ""      }, { "c",   "0"    } }, {             } },
        { "a=&b=0;c=0",  I,         { { "a",     ""     }, { "b",    "0"     }, { "c",   "0"    } }, {             } },
        { "a=0&;",       I,         { { "a",     "0"    },                                        }, {             } },
        { "a=0&b;",      I,         { { "a",     "0"    }, { "b",    ""      },                   }, {             } },
        { "a=0&b=;",     I,         { { "a",     "0"    }, { "b",    ""      },                   }, {             } },
        { "a=0&b=0;",    I,         { { "a",     "0"    }, { "b",    "0"     },                   }, {             } },
        { "a=0&;c",      I,         { { "a",     "0"    },                      { "c",   ""     } }, {             } },
        { "a=0&b;c",     I,         { { "a",     "0"    }, { "b",    ""      }, { "c",   ""     } }, {             } },
        { "a=0&b=;c",    I,         { { "a",     "0"    }, { "b",    ""      }, { "c",   ""     } }, {             } },
        { "a=0&b=0;c",   I,         { { "a",     "0"    }, { "b",    "0"     }, { "c",   ""     } }, {             } },
        { "a=0&;c=",     I,         { { "a",     "0"    },                      { "c",   ""     } }, {             } },
        { "a=0&b;c=",    I,         { { "a",     "0"    }, { "b",    ""      }, { "c",   ""     } }, {             } },
        { "a=0&b=;c=",   I,         { { "a",     "0"    }, { "b",    ""      }, { "c",   ""     } }, {             } },
        { "a=0&b=0;c=",  I,         { { "a",     "0"    }, { "b",    "0"     }, { "c",   ""     } }, {             } },
        { "a=0&;c=0",    I,         { { "a",     "0"    },                      { "c",   "0"    } }, {             } },
        { "a=0&b;c=0",   I,         { { "a",     "0"    }, { "b",    ""      }, { "c",   "0"    } }, {             } },
        { "a=0&b=;c=0",  I,         { { "a",     "0"    }, { "b",    ""      }, { "c",   "0"    } }, {             } },
        { "a=0&b=0;c=0", I,         { { "a",     "0"    }, { "b",    "0"     }, { "c",   "0"    } }, {             } },
        { "",            N | I,     {                                                             }, {             } },
        { "a=",          N | I,     { { "a",     ""     },                                        }, {             } },
        { "a=0",         N | I,     { { "a",     "0"    },                                        }, {             } },
        { "&",           N | I,     {                                                             }, {             } },
        { "a&",          N | I,     { { "a",     ""     },                                        }, {             } },
        { "a=&",         N | I,     { { "a",     ""     },                                        }, {             } },
        { "a=0&",        N | I,     { { "a",     "0"    },                                        }, {             } },
        { "&b",          N | I,     {                      { "b",    ""      },                   }, {             } },
        { "a&b",         N | I,     { { "a",     ""     }, { "b",    ""      },                   }, {             } },
        { "a=&b",        N | I,     { { "a",     ""     }, { "b",    ""      },                   }, {             } },
        { "a=0&b",       N | I,     { { "a",     "0"    }, { "b",    ""      },                   }, {             } },
        { "&b=",         N | I,     {                      { "b",    ""      },                   }, {             } },
        { "a&b=",        N | I,     { { "a",     ""     }, { "b",    ""      },                   }, {             } },
        { "a=&b=",       N | I,     { { "a",     ""     }, { "b",    ""      },                   }, {             } },
        { "a=0&b=",      N | I,     { { "a",     "0"    }, { "b",    ""      },                   }, {             } },
        { "&b=0",        N | I,     {                      { "b",    "0"     },                   }, {             } },
        { "a&b=0",       N | I,     { { "a",     ""     }, { "b",    "0"     },                   }, {             } },
        { "a=&b=0",      N | I,     { { "a",     ""     }, { "b",    "0"     },                   }, {             } },
        { "a=0&b=0",     N | I,     { { "a",     "0"    }, { "b",    "0"     },                   }, {             } },
        { ";",           N | I,     {                                                             }, {             } },
        { "b;",          N | I,     {                      { "b",    ""      },                   }, {             } },
        { "b=;",         N | I,     {                      { "b",    ""      },                   }, {             } },
        { "b=0;",        N | I,     {                      { "b",    "0"     },                   }, {             } },
        { ";c",          N | I,     {                                           { "c",   ""     } }, {             } },
        { "b;c",         N | I,     {                      { "b",    ""      }, { "c",   ""     } }, {             } },
        { "b=;c",        N | I,     {                      { "b",    ""      }, { "c",   ""     } }, {             } },
        { "b=0;c",       N | I,     {                      { "b",    "0"     }, { "c",   ""     } }, {             } },
        { ";c=",         N | I,     {                                           { "c",   ""     } }, {             } },
        { "b;c=",        N | I,     {                      { "b",    ""      }, { "c",   ""     } }, {             } },
        { "b=;c=",       N | I,     {                      { "b",    ""      }, { "c",   ""     } }, {             } },
        { "b=0;c=",      N | I,     {                      { "b",    "0"     }, { "c",   ""     } }, {             } },
        { ";c=0",        N | I,     {                                           { "c",   "0"    } }, {             } },
        { "b;c=0",       N | I,     {                      { "b",    ""      }, { "c",   "0"    } }, {             } },
        { "b=;c=0",      N | I,     {                      { "b",    ""      }, { "c",   "0"    } }, {             } },
        { "b=0;c=0",     N | I,     {                      { "b",    "0"     }, { "c",   "0"    } }, {             } },
        { "&;",          N | I,     {                                                             }, {             } },
        { "&b;",         N | I,     {                      { "b",    ""      },                   }, {             } },
        { "&b=;",        N | I,     {                      { "b",    ""      },                   }, {             } },
        { "&b=0;",       N | I,     {                      { "b",    "0"     },                   }, {             } },
        { "&;c",         N | I,     {                                           { "c",   ""     } }, {             } },
        { "&b;c",        N | I,     {                      { "b",    ""      }, { "c",   ""     } }, {             } },
        { "&b=;c",       N | I,     {                      { "b",    ""      }, { "c",   ""     } }, {             } },
        { "&b=0;c",      N | I,     {                      { "b",    "0"     }, { "c",   ""     } }, {             } },
        { "&;c=",        N | I,     {                                           { "c",   ""     } }, {             } },
        { "&b;c=",       N | I,     {                      { "b",    ""      }, { "c",   ""     } }, {             } },
        { "&b=;c=",      N | I,     {                      { "b",    ""      }, { "c",   ""     } }, {             } },
        { "&b=0;c=",     N | I,     {                      { "b",    "0"     }, { "c",   ""     } }, {             } },
        { "&;c=0",       N | I,     {                                           { "c",   "0"    } }, {             } },
        { "&b;c=0",      N | I,     {                      { "b",    ""      }, { "c",   "0"    } }, {             } },
        { "&b=;c=0",     N | I,     {                      { "b",    ""      }, { "c",   "0"    } }, {             } },
        { "&b=0;c=0",    N | I,     {                      { "b",    "0"     }, { "c",   "0"    } }, {             } },
        { "a&;",         N | I,     { { "a",     ""     },                                        }, {             } },
        { "a&b;",        N | I,     { { "a",     ""     }, { "b",    ""      },                   }, {             } },
        { "a&b=;",       N | I,     { { "a",     ""     }, { "b",    ""      },                   }, {             } },
        { "a&b=0;",      N | I,     { { "a",     ""     }, { "b",    "0"     },                   }, {             } },
        { "a&;c",        N | I,     { { "a",     ""     },                      { "c",   ""     } }, {             } },
        { "a&b;c",       N | I,     { { "a",     ""     }, { "b",    ""      }, { "c",   ""     } }, {             } },
        { "a&b=;c",      N | I,     { { "a",     ""     }, { "b",    ""      }, { "c",   ""     } }, {             } },
        { "a&b=0;c",     N | I,     { { "a",     ""     }, { "b",    "0"     }, { "c",   ""     } }, {             } },
        { "a&;c=",       N | I,     { { "a",     ""     },                      { "c",   ""     } }, {             } },
        { "a&b;c=",      N | I,     { { "a",     ""     }, { "b",    ""      }, { "c",   ""     } }, {             } },
        { "a&b=;c=",     N | I,     { { "a",     ""     }, { "b",    ""      }, { "c",   ""     } }, {             } },
        { "a&b=0;c=",    N | I,     { { "a",     ""     }, { "b",    "0"     }, { "c",   ""     } }, {             } },
        { "a&;c=0",      N | I,     { { "a",     ""     },                      { "c",   "0"    } }, {             } },
        { "a&b;c=0",     N | I,     { { "a",     ""     }, { "b",    ""      }, { "c",   "0"    } }, {             } },
        { "a&b=;c=0",    N | I,     { { "a",     ""     }, { "b",    ""      }, { "c",   "0"    } }, {             } },
        { "a&b=0;c=0",   N | I,     { { "a",     ""     }, { "b",    "0"     }, { "c",   "0"    } }, {             } },
        { "a=&;",        N | I,     { { "a",     ""     },                                        }, {             } },
        { "a=&b;",       N | I,     { { "a",     ""     }, { "b",    ""      },                   }, {             } },
        { "a=&b=;",      N | I,     { { "a",     ""     }, { "b",    ""      },                   }, {             } },
        { "a=&b=0;",     N | I,     { { "a",     ""     }, { "b",    "0"     },                   }, {             } },
        { "a=&;c",       N | I,     { { "a",     ""     },                      { "c",   ""     } }, {             } },
        { "a=&b;c",      N | I,     { { "a",     ""     }, { "b",    ""      }, { "c",   ""     } }, {             } },
        { "a=&b=;c",     N | I,     { { "a",     ""     }, { "b",    ""      }, { "c",   ""     } }, {             } },
        { "a=&b=0;c",    N | I,     { { "a",     ""     }, { "b",    "0"     }, { "c",   ""     } }, {             } },
        { "a=&;c=",      N | I,     { { "a",     ""     },                      { "c",   ""     } }, {             } },
        { "a=&b;c=",     N | I,     { { "a",     ""     }, { "b",    ""      }, { "c",   ""     } }, {             } },
        { "a=&b=;c=",    N | I,     { { "a",     ""     }, { "b",    ""      }, { "c",   ""     } }, {             } },
        { "a=&b=0;c=",   N | I,     { { "a",     ""     }, { "b",    "0"     }, { "c",   ""     } }, {             } },
        { "a=&;c=0",     N | I,     { { "a",     ""     },                      { "c",   "0"    } }, {             } },
        { "a=&b;c=0",    N | I,     { { "a",     ""     }, { "b",    ""      }, { "c",   "0"    } }, {             } },
        { "a=&b=;c=0",   N | I,     { { "a",     ""     }, { "b",    ""      }, { "c",   "0"    } }, {             } },
        { "a=&b=0;c=0",  N | I,     { { "a",     ""     }, { "b",    "0"     }, { "c",   "0"    } }, {             } },
        { "a=0&;",       N | I,     { { "a",     "0"    },                                        }, {             } },
        { "a=0&b;",      N | I,     { { "a",     "0"    }, { "b",    ""      },                   }, {             } },
        { "a=0&b=;",     N | I,     { { "a",     "0"    }, { "b",    ""      },                   }, {             } },
        { "a=0&b=0;",    N | I,     { { "a",     "0"    }, { "b",    "0"     },                   }, {             } },
        { "a=0&;c",      N | I,     { { "a",     "0"    },                      { "c",   ""     } }, {             } },
        { "a=0&b;c",     N | I,     { { "a",     "0"    }, { "b",    ""      }, { "c",   ""     } }, {             } },
        { "a=0&b=;c",    N | I,     { { "a",     "0"    }, { "b",    ""      }, { "c",   ""     } }, {             } },
        { "a=0&b=0;c",   N | I,     { { "a",     "0"    }, { "b",    "0"     }, { "c",   ""     } }, {             } },
        { "a=0&;c=",     N | I,     { { "a",     "0"    },                      { "c",   ""     } }, {             } },
        { "a=0&b;c=",    N | I,     { { "a",     "0"    }, { "b",    ""      }, { "c",   ""     } }, {             } },
        { "a=0&b=;c=",   N | I,     { { "a",     "0"    }, { "b",    ""      }, { "c",   ""     } }, {             } },
        { "a=0&b=0;c=",  N | I,     { { "a",     "0"    }, { "b",    "0"     }, { "c",   ""     } }, {             } },
        { "a=0&;c=0",    N | I,     { { "a",     "0"    },                      { "c",   "0"    } }, {             } },
        { "a=0&b;c=0",   N | I,     { { "a",     "0"    }, { "b",    ""      }, { "c",   "0"    } }, {             } },
        { "a=0&b=;c=0",  N | I,     { { "a",     "0"    }, { "b",    ""      }, { "c",   "0"    } }, {             } },
        { "a=0&b=0;c=0", N | I,     { { "a",     "0"    }, { "b",    "0"     }, { "c",   "0"    } }, {             } },
        { "",            S | I,     {                                                             }, {             } },
        { "a=",          S | I,     { { "a",     ""     },                                        }, {             } },
        { "a=0",         S | I,     { { "a",     "0"    },                                        }, {             } },
        { "&",           S | I,     {                                                             }, {             } },
        { "a&",          S | I,     { { "a",     ""     },                                        }, {             } },
        { "a=&",         S | I,     { { "a",     ""     },                                        }, {             } },
        { "a=0&",        S | I,     { { "a",     "0"    },                                        }, {             } },
        { "&b",          S | I,     {                      { "b",    ""      },                   }, {             } },
        { "a&b",         S | I,     { { "a",     ""     }, { "b",    ""      },                   }, {             } },
        { "a=&b",        S | I,     { { "a",     ""     }, { "b",    ""      },                   }, {             } },
        { "a=0&b",       S | I,     { { "a",     "0"    }, { "b",    ""      },                   }, {             } },
        { "&b=",         S | I,     {                      { "b",    ""      },                   }, {             } },
        { "a&b=",        S | I,     { { "a",     ""     }, { "b",    ""      },                   }, {             } },
        { "a=&b=",       S | I,     { { "a",     ""     }, { "b",    ""      },                   }, {             } },
        { "a=0&b=",      S | I,     { { "a",     "0"    }, { "b",    ""      },                   }, {             } },
        { "&b=0",        S | I,     {                      { "b",    "0"     },                   }, {             } },
        { "a&b=0",       S | I,     { { "a",     ""     }, { "b",    "0"     },                   }, {             } },
        { "a=&b=0",      S | I,     { { "a",     ""     }, { "b",    "0"     },                   }, {             } },
        { "a=0&b=0",     S | I,     { { "a",     "0"    }, { "b",    "0"     },                   }, {             } },
        { ";",           S | I,     {                      { ";",    ""      },                   }, {             } },
        { "b;",          S | I,     {                      { "b;",   ""      },                   }, {             } },
        { "b=;",         S | I,     {                      { "b",    ";"     },                   }, {             } },
        { "b=0;",        S | I,     {                      { "b",    "0;"    },                   }, {             } },
        { ";c",          S | I,     {                                           { ";c",  ""     } }, {             } },
        { "b;c",         S | I,     {                      { "b;c",  ""      },                   }, {             } },
        { "b=;c",        S | I,     {                      { "b",    ";c"    },                   }, {             } },
        { "b=0;c",       S | I,     {                      { "b",    "0;c"   },                   }, {             } },
        { ";c=",         S | I,     {                                           { ";c",  ""     } }, {             } },
        { "b;c=",        S | I,     {                      { "b;c",  ""      }                    }, {             } },
        { "b=;c=",       S | I,     {                      { "b",    ";c="   },                   }, {             } },
        { "b=0;c=",      S | I,     {                      { "b",    "0;c="  },                   }, {             } },
        { ";c=0",        S | I,     {                                           { ";c",  "0"    } }, {             } },
        { "b;c=0",       S | I,     {                      { "b;c",  "0"     },                   }, {             } },
        { "b=;c=0",      S | I,     {                      { "b",    ";c=0"  },                   }, {             } },
        { "b=0;c=0",     S | I,     {                      { "b",    "0;c=0" },                   }, {             } },
        { "&;",          S | I,     {                      { ";",    ""      },                   }, {             } },
        { "&b;",         S | I,     {                      { "b;",   ""      },                   }, {             } },
        { "&b=;",        S | I,     {                      { "b",    ";"     },                   }, {             } },
        { "&b=0;",       S | I,     {                      { "b",    "0;"    },                   }, {             } },
        { "&;c",         S | I,     {                                           { ";c",  ""     } }, {             } },
        { "&b;c",        S | I,     {                      { "b;c",  ""      },                   }, {             } },
        { "&b=;c",       S | I,     {                      { "b",    ";c"    },                   }, {             } },
        { "&b=0;c",      S | I,     {                      { "b",    "0;c"   },                   }, {             } },
        { "&;c=",        S | I,     {                                           { ";c",  ""     } }, {             } },
        { "&b;c=",       S | I,     {                      { "b;c",  ""      },                   }, {             } },
        { "&b=;c=",      S | I,     {                      { "b",    ";c="   },                   }, {             } },
        { "&b=0;c=",     S | I,     {                      { "b",    "0;c="  },                   }, {             } },
        { "&;c=0",       S | I,     {                                           { ";c",  "0"    } }, {             } },
        { "&b;c=0",      S | I,     {                      { "b;c",  "0"     },                   }, {             } },
        { "&b=;c=0",     S | I,     {                      { "b",    ";c=0"  },                   }, {             } },
        { "&b=0;c=0",    S | I,     {                      { "b",    "0;c=0" },                   }, {             } },
        { "a&;",         S | I,     { { "a",     ""     }, { ";",    ""      },                   }, {             } },
        { "a&b;",        S | I,     { { "a",     ""     }, { "b;",   ""      },                   }, {             } },
        { "a&b=;",       S | I,     { { "a",     ""     }, { "b",    ";"     },                   }, {             } },
        { "a&b=0;",      S | I,     { { "a",     ""     }, { "b",    "0;"    },                   }, {             } },
        { "a&;c",        S | I,     { { "a",     ""     },                      { ";c",  ""     } }, {             } },
        { "a&b;c",       S | I,     { { "a",     ""     }, { "b;c",  ""      },                   }, {             } },
        { "a&b=;c",      S | I,     { { "a",     ""     }, { "b",    ";c"    },                   }, {             } },
        { "a&b=0;c",     S | I,     { { "a",     ""     }, { "b",    "0;c"   },                   }, {             } },
        { "a&;c=",       S | I,     { { "a",     ""     },                      { ";c",  ""     } }, {             } },
        { "a&b;c=",      S | I,     { { "a",     ""     }, { "b;c",  ""      },                   }, {             } },
        { "a&b=;c=",     S | I,     { { "a",     ""     }, { "b",    ";c="   },                   }, {             } },
        { "a&b=0;c=",    S | I,     { { "a",     ""     }, { "b",    "0;c="  },                   }, {             } },
        { "a&;c=0",      S | I,     { { "a",     ""     },                      { ";c",  "0"    } }, {             } },
        { "a&b;c=0",     S | I,     { { "a",     ""     }, { "b;c",  "0"     },                   }, {             } },
        { "a&b=;c=0",    S | I,     { { "a",     ""     }, { "b",    ";c=0"  },                   }, {             } },
        { "a&b=0;c=0",   S | I,     { { "a",     ""     }, { "b",    "0;c=0" },                   }, {             } },
        { "a=&;",        S | I,     { { "a",     ""     }, { ";",    ""      },                   }, {             } },
        { "a=&b;",       S | I,     { { "a",     ""     }, { "b;",   ""      },                   }, {             } },
        { "a=&b=;",      S | I,     { { "a",     ""     }, { "b",    ";"     },                   }, {             } },
        { "a=&b=0;",     S | I,     { { "a",     ""     }, { "b",    "0;"    },                   }, {             } },
        { "a=&;c",       S | I,     { { "a",     ""     },                      { ";c",  ""     } }, {             } },
        { "a=&b;c",      S | I,     { { "a",     ""     }, { "b;c",  ""      },                   }, {             } },
        { "a=&b=;c",     S | I,     { { "a",     ""     }, { "b",    ";c"    },                   }, {             } },
        { "a=&b=0;c",    S | I,     { { "a",     ""     }, { "b",    "0;c"   },                   }, {             } },
        { "a=&;c=",      S | I,     { { "a",     ""     },                      { ";c",  ""     } }, {             } },
        { "a=&b;c=",     S | I,     { { "a",     ""     }, { "b;c",  ""      },                   }, {             } },
        { "a=&b=;c=",    S | I,     { { "a",     ""     }, { "b",    ";c="   },                   }, {             } },
        { "a=&b=0;c=",   S | I,     { { "a",     ""     }, { "b",    "0;c="  },                   }, {             } },
        { "a=&;c=0",     S | I,     { { "a",     ""     },                      { ";c",  "0"    } }, {             } },
        { "a=&b;c=0",    S | I,     { { "a",     ""     }, { "b;c",  "0"     },                   }, {             } },
        { "a=&b=;c=0",   S | I,     { { "a",     ""     }, { "b",    ";c=0"  },                   }, {             } },
        { "a=&b=0;c=0",  S | I,     { { "a",     ""     }, { "b",    "0;c=0" },                   }, {             } },
        { "a=0&;",       S | I,     { { "a",     "0"    }, { ";",    ""      },                   }, {             } },
        { "a=0&b;",      S | I,     { { "a",     "0"    }, { "b;",   ""      },                   }, {             } },
        { "a=0&b=;",     S | I,     { { "a",     "0"    }, { "b",    ";"     },                   }, {             } },
        { "a=0&b=0;",    S | I,     { { "a",     "0"    }, { "b",    "0;"    },                   }, {             } },
        { "a=0&;c",      S | I,     { { "a",     "0"    },                      { ";c",  ""     } }, {             } },
        { "a=0&b;c",     S | I,     { { "a",     "0"    }, { "b;c",  ""      },                   }, {             } },
        { "a=0&b=;c",    S | I,     { { "a",     "0"    }, { "b",    ";c"    },                   }, {             } },
        { "a=0&b=0;c",   S | I,     { { "a",     "0"    }, { "b",    "0;c"   },                   }, {             } },
        { "a=0&;c=",     S | I,     { { "a",     "0"    },                      { ";c",  ""     } }, {             } },
        { "a=0&b;c=",    S | I,     { { "a",     "0"    }, { "b;c",  ""      },                   }, {             } },
        { "a=0&b=;c=",   S | I,     { { "a",     "0"    }, { "b",    ";c="   },                   }, {             } },
        { "a=0&b=0;c=",  S | I,     { { "a",     "0"    }, { "b",    "0;c="  },                   }, {             } },
        { "a=0&;c=0",    S | I,     { { "a",     "0"    },                      { ";c",  "0"    } }, {             } },
        { "a=0&b;c=0",   S | I,     { { "a",     "0"    }, { "b;c",  "0"     },                   }, {             } },
        { "a=0&b=;c=0",  S | I,     { { "a",     "0"    }, { "b",    ";c=0"  },                   }, {             } },
        { "a=0&b=0;c=0", S | I,     { { "a",     "0"    }, { "b",    "0;c=0" },                   }, {             } },
        { "",            N | S | I, {                                                             }, {             } },
        { "a=",          N | S | I, { { "a",     ""     },                                        }, {             } },
        { "a=0",         N | S | I, { { "a",     "0"    },                                        }, {             } },
        { "&",           N | S | I, {                                                             }, {             } },
        { "a&",          N | S | I, { { "a",     ""     },                                        }, {             } },
        { "a=&",         N | S | I, { { "a",     ""     },                                        }, {             } },
        { "a=0&",        N | S | I, { { "a",     "0"    },                                        }, {             } },
        { "&b",          N | S | I, {                      { "b",    ""      },                   }, {             } },
        { "a&b",         N | S | I, { { "a",     ""     }, { "b",    ""      },                   }, {             } },
        { "a=&b",        N | S | I, { { "a",     ""     }, { "b",    ""      },                   }, {             } },
        { "a=0&b",       N | S | I, { { "a",     "0"    }, { "b",    ""      },                   }, {             } },
        { "&b=",         N | S | I, {                      { "b",    ""      },                   }, {             } },
        { "a&b=",        N | S | I, { { "a",     ""     }, { "b",    ""      },                   }, {             } },
        { "a=&b=",       N | S | I, { { "a",     ""     }, { "b",    ""      },                   }, {             } },
        { "a=0&b=",      N | S | I, { { "a",     "0"    }, { "b",    ""      },                   }, {             } },
        { "&b=0",        N | S | I, {                      { "b",    "0"     },                   }, {             } },
        { "a&b=0",       N | S | I, { { "a",     ""     }, { "b",    "0"     },                   }, {             } },
        { "a=&b=0",      N | S | I, { { "a",     ""     }, { "b",    "0"     },                   }, {             } },
        { "a=0&b=0",     N | S | I, { { "a",     "0"    }, { "b",    "0"     },                   }, {             } },
        { ";",           N | S | I, {                      { ";",    ""      },                   }, {             } },
        { "b;",          N | S | I, {                      { "b;",   ""      },                   }, {             } },
        { "b=;",         N | S | I, {                      { "b",    ";"     },                   }, {             } },
        { "b=0;",        N | S | I, {                      { "b",    "0;"    },                   }, {             } },
        { ";c",          N | S | I, {                                           { ";c",  ""     } }, {             } },
        { "b;c",         N | S | I, {                      { "b;c",  ""      },                   }, {             } },
        { "b=;c",        N | S | I, {                      { "b",    ";c"    },                   }, {             } },
        { "b=0;c",       N | S | I, {                      { "b",    "0;c"   },                   }, {             } },
        { ";c=",         N | S | I, {                                           { ";c",  ""     } }, {             } },
        { "b;c=",        N | S | I, {                      { "b;c",  ""      }                    }, {             } },
        { "b=;c=",       N | S | I, {                      { "b",    ";c="   },                   }, {             } },
        { "b=0;c=",      N | S | I, {                      { "b",    "0;c="  },                   }, {             } },
        { ";c=0",        N | S | I, {                                           { ";c",  "0"    } }, {             } },
        { "b;c=0",       N | S | I, {                      { "b;c",  "0"     },                   }, {             } },
        { "b=;c=0",      N | S | I, {                      { "b",    ";c=0"  },                   }, {             } },
        { "b=0;c=0",     N | S | I, {                      { "b",    "0;c=0" },                   }, {             } },
        { "&;",          N | S | I, {                      { ";",    ""      },                   }, {             } },
        { "&b;",         N | S | I, {                      { "b;",   ""      },                   }, {             } },
        { "&b=;",        N | S | I, {                      { "b",    ";"     },                   }, {             } },
        { "&b=0;",       N | S | I, {                      { "b",    "0;"    },                   }, {             } },
        { "&;c",         N | S | I, {                                           { ";c",  ""     } }, {             } },
        { "&b;c",        N | S | I, {                      { "b;c",  ""      },                   }, {             } },
        { "&b=;c",       N | S | I, {                      { "b",    ";c"    },                   }, {             } },
        { "&b=0;c",      N | S | I, {                      { "b",    "0;c"   },                   }, {             } },
        { "&;c=",        N | S | I, {                                           { ";c",  ""     } }, {             } },
        { "&b;c=",       N | S | I, {                      { "b;c",  ""      },                   }, {             } },
        { "&b=;c=",      N | S | I, {                      { "b",    ";c="   },                   }, {             } },
        { "&b=0;c=",     N | S | I, {                      { "b",    "0;c="  },                   }, {             } },
        { "&;c=0",       N | S | I, {                                           { ";c",  "0"    } }, {             } },
        { "&b;c=0",      N | S | I, {                      { "b;c",  "0"     },                   }, {             } },
        { "&b=;c=0",     N | S | I, {                      { "b",    ";c=0"  },                   }, {             } },
        { "&b=0;c=0",    N | S | I, {                      { "b",    "0;c=0" },                   }, {             } },
        { "a&;",         N | S | I, { { "a",     ""     }, { ";",    ""      },                   }, {             } },
        { "a&b;",        N | S | I, { { "a",     ""     }, { "b;",   ""      },                   }, {             } },
        { "a&b=;",       N | S | I, { { "a",     ""     }, { "b",    ";"     },                   }, {             } },
        { "a&b=0;",      N | S | I, { { "a",     ""     }, { "b",    "0;"    },                   }, {             } },
        { "a&;c",        N | S | I, { { "a",     ""     },                      { ";c",  ""     } }, {             } },
        { "a&b;c",       N | S | I, { { "a",     ""     }, { "b;c",  ""      },                   }, {             } },
        { "a&b=;c",      N | S | I, { { "a",     ""     }, { "b",    ";c"    },                   }, {             } },
        { "a&b=0;c",     N | S | I, { { "a",     ""     }, { "b",    "0;c"   },                   }, {             } },
        { "a&;c=",       N | S | I, { { "a",     ""     },                      { ";c",  ""     } }, {             } },
        { "a&b;c=",      N | S | I, { { "a",     ""     }, { "b;c",  ""      },                   }, {             } },
        { "a&b=;c=",     N | S | I, { { "a",     ""     }, { "b",    ";c="   },                   }, {             } },
        { "a&b=0;c=",    N | S | I, { { "a",     ""     }, { "b",    "0;c="  },                   }, {             } },
        { "a&;c=0",      N | S | I, { { "a",     ""     },                      { ";c",  "0"    } }, {             } },
        { "a&b;c=0",     N | S | I, { { "a",     ""     }, { "b;c",  "0"     },                   }, {             } },
        { "a&b=;c=0",    N | S | I, { { "a",     ""     }, { "b",    ";c=0"  },                   }, {             } },
        { "a&b=0;c=0",   N | S | I, { { "a",     ""     }, { "b",    "0;c=0" },                   }, {             } },
        { "a=&;",        N | S | I, { { "a",     ""     }, { ";",    ""      },                   }, {             } },
        { "a=&b;",       N | S | I, { { "a",     ""     }, { "b;",   ""      },                   }, {             } },
        { "a=&b=;",      N | S | I, { { "a",     ""     }, { "b",    ";"     },                   }, {             } },
        { "a=&b=0;",     N | S | I, { { "a",     ""     }, { "b",    "0;"    },                   }, {             } },
        { "a=&;c",       N | S | I, { { "a",     ""     },                      { ";c",  ""     } }, {             } },
        { "a=&b;c",      N | S | I, { { "a",     ""     }, { "b;c",  ""      },                   }, {             } },
        { "a=&b=;c",     N | S | I, { { "a",     ""     }, { "b",    ";c"    },                   }, {             } },
        { "a=&b=0;c",    N | S | I, { { "a",     ""     }, { "b",    "0;c"   },                   }, {             } },
        { "a=&;c=",      N | S | I, { { "a",     ""     },                      { ";c",  ""     } }, {             } },
        { "a=&b;c=",     N | S | I, { { "a",     ""     }, { "b;c",  ""      },                   }, {             } },
        { "a=&b=;c=",    N | S | I, { { "a",     ""     }, { "b",    ";c="   },                   }, {             } },
        { "a=&b=0;c=",   N | S | I, { { "a",     ""     }, { "b",    "0;c="  },                   }, {             } },
        { "a=&;c=0",     N | S | I, { { "a",     ""     },                      { ";c",  "0"    } }, {             } },
        { "a=&b;c=0",    N | S | I, { { "a",     ""     }, { "b;c",  "0"     },                   }, {             } },
        { "a=&b=;c=0",   N | S | I, { { "a",     ""     }, { "b",    ";c=0"  },                   }, {             } },
        { "a=&b=0;c=0",  N | S | I, { { "a",     ""     }, { "b",    "0;c=0" },                   }, {             } },
        { "a=0&;",       N | S | I, { { "a",     "0"    }, { ";",    ""      },                   }, {             } },
        { "a=0&b;",      N | S | I, { { "a",     "0"    }, { "b;",   ""      },                   }, {             } },
        { "a=0&b=;",     N | S | I, { { "a",     "0"    }, { "b",    ";"     },                   }, {             } },
        { "a=0&b=0;",    N | S | I, { { "a",     "0"    }, { "b",    "0;"    },                   }, {             } },
        { "a=0&;c",      N | S | I, { { "a",     "0"    },                      { ";c",  ""     } }, {             } },
        { "a=0&b;c",     N | S | I, { { "a",     "0"    }, { "b;c",  ""      },                   }, {             } },
        { "a=0&b=;c",    N | S | I, { { "a",     "0"    }, { "b",    ";c"    },                   }, {             } },
        { "a=0&b=0;c",   N | S | I, { { "a",     "0"    }, { "b",    "0;c"   },                   }, {             } },
        { "a=0&;c=",     N | S | I, { { "a",     "0"    },                      { ";c",  ""     } }, {             } },
        { "a=0&b;c=",    N | S | I, { { "a",     "0"    }, { "b;c",  ""      },                   }, {             } },
        { "a=0&b=;c=",   N | S | I, { { "a",     "0"    }, { "b",    ";c="   },                   }, {             } },
        { "a=0&b=0;c=",  N | S | I, { { "a",     "0"    }, { "b",    "0;c="  },                   }, {             } },
        { "a=0&;c=0",    N | S | I, { { "a",     "0"    },                      { ";c",  "0"    } }, {             } },
        { "a=0&b;c=0",   N | S | I, { { "a",     "0"    }, { "b;c",  "0"     },                   }, {             } },
        { "a=0&b=;c=0",  N | S | I, { { "a",     "0"    }, { "b",    ";c=0"  },                   }, {             } },
        { "a=0&b=0;c=0", N | S | I, { { "a",     "0"    }, { "b",    "0;c=0" },                   }, {             } },
    };

    for (const auto& check : checks) {
        const auto& source = get<eQuery>(check);
        const auto flags = get<eFlags>(check);
        const auto& expected_entries = get<eExpectedEntries>(check);
        const auto& expected_indexes = get<eExpectedIndexes>(check);

        TCgiEntries parsed_entries;
        TCgiIndexes parsed_indexes;
        CCgiEntries_Parser parser(&parsed_entries, &parsed_indexes, flags);
        parser.SetQueryString(source);

        if (parsed_entries.size() != expected_entries.size()) {
            BOOST_ERROR("Parsed entries size (" << parsed_entries.size() << ") do not match expected ones (" <<
                    expected_entries.size() << ") for '" << source << "' and flags=" << flags);
        } else if (parsed_indexes.size() != expected_indexes.size()) {
            BOOST_ERROR("Parsed indexes size (" << parsed_indexes.size() << ") do not match expected ones (" <<
                    expected_indexes.size() << ") for '" << source << "' and flags=" << flags);
        } else {
            auto r1 = mismatch(expected_entries.begin(), expected_entries.end(), parsed_entries.begin());

            if (r1.first != expected_entries.end()) {
                BOOST_ERROR("Parsed entries does not match expected ones, first mismatch: '" <<
                        r1.second->first << "'='" << r1.second->second.GetValue() << "' vs '" <<
                        r1.first->first << "'='" << r1.first->second.GetValue() << "' for '" << source << "' and flags=" << flags);
            } else {
                auto r2 = mismatch(expected_indexes.begin(), expected_indexes.end(), parsed_indexes.begin());

                if (r2.first != expected_indexes.end()) {
                    BOOST_ERROR("Parsed indexes does not match expected ones, first mismatch: '" <<
                            *r2.second << "' vs '" << *r2.first << "' for '" << source << "' and flags=" << flags);
                }
            }
        }
    }
}
